var prefRoot = "extensions.vcssupport.";
var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(prefRoot);
var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
var windowWatcher = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);

function calVcsExporter() {
	this.wrappedJSObject = this;
}

calVcsExporter.prototype.QueryInterface = function QueryInterface(aIID) {
	if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.calIExporter)) {
		throw Components.results.NS_ERROR_NO_INTERFACE;
	}
	return this;
};

calVcsExporter.prototype.getFileTypes = function getFileTypes(aCount) {
	aCount.value = 1;
	return([{defaultExtension:'vcs', extensionFilter:'*.vcs', description:stringName("vcsFileType")}]);
};

calVcsExporter.prototype.exportToStream = function vcs_exportToStream(aStream, aCount, aItems) {

	//Open the options window (and don't let it do anything until the options window is closed so that we can set our prefs before the export happens) 
	//A small bug is that we have already opened the file to stream data into it, of it nothing
	windowWatcher.openWindow(null, "chrome://vcssupport/content/vcssupportOptions.xul",	null, "chrome, centerscreen, modal", null);
						
	var str = "";
	var headers = [];
	// Not using a loop here, since we need to be sure the order here matches
	// with the orders the field data is added later on

	// The structure is STRUCTURE (HEADER) ---  INFO  (EVENTS) --- STRUCTURE (FOOTER)

	headers.push("BEGIN:VCALENDAR");
	headers.push("VERSION:1.0");
	headers.push("PROD-ID:-//Mozilla.org/NONSGML Mozilla Calendar V1.1//EN");
	
	var beginning = headers.join("\r\n")+"\r\n"
		
	//export_recurrences_as_rules - Decides whether to use RRULE or individual events
	var rulePref = prefManager.getBoolPref("export_recurrences_as_rules");

	var items = "";
	
	for each (item in aItems) {
		var line = [];

		if (rulePref) {
			if (item.recurrenceInfo) {
				line.push(recurringEventRules(item));  //for items with recurrences where we want rules exported
			} else {
				line.push(singleEventRules(item));  //for single items where we want rules exported
			}
		} else {
			if (item.recurrenceInfo) {
				line.push(recurringEventItems(item));  //for items with recurrences where we want occurrences exported
			} else {
				line.push(singleEventItems(item));  //for single items where we want occurrences exported
			}
		}

		items = items + line.join("");
	}

	var footers = [];
	footers.push("END:VCALENDAR");

	var ending = footers.join("\r\n")+"\r\n";
	
	str = beginning + items + ending;
	
	//If the length of the string to be written is 103 characters, then no events are in the string to be written (just the header and footer).  Show a warning in this case
	if (str.length <= 103) {
		promptService.alert(null, stringName("noItemsTitle") , stringName("noItemsWarning"));
	} else {
		aStream.write(str, str.length);
	}
	
	return;
}

//Helper functions

function singleEventRules(item) {
	var eventType = "single";
	var line = [];
	var alarmOffset = item.alarmOffset;
	line.push(serializeEvent(item, eventType, alarmOffset));
	details = line.join("");
	return details;
}

function singleEventItems(item) {
	var eventType = "single";
	var alarmOffset = item.alarmOffset;
	var line = [];
	
	var date;
	if (item.startDate) { 
		date = item.startDate.nativeTime;
	} else if (item.entryDate) { 
		date = item.entryDate.nativeTime;
	} else { 
		date = 0;
	}
	var startDiff = prefManager.getIntPref("export_start_date")*100000000 - date;
	
	if (item.endDate) { 
		date = item.endDate.nativeTime;
	} else if (item.dueDate) { 
		date = item.dueDate.nativeTime;
	} else { 
		date = 0
	}
	var endDiff = prefManager.getIntPref("export_end_date")*100000000 - date;
	
	//Get the details about each item between startDate and endDate
	if (startDiff < 0 && endDiff > 0) {
		line.push(serializeEvent(item, eventType, alarmOffset));
	}
	details = line.join("");
	return details;
}

function recurringEventRules(item) {
	var eventType = "single";
	var line = [];
	var alarmOffset = item.alarmOffset;
	line.push(serializeEvent(item, eventType, alarmOffset));
	details = line.join("");
	return details;
}

function recurringEventItems(item) {
	var eventType = "multiple";
	//Define the start time and the end time from the prefs
	var startDate = new Date();
	startDate.setTime(prefManager.getIntPref("export_start_date")*100000);
	startDate = jsDateToDateTime(startDate);
	var endDate = new Date();
	endDate.setTime(prefManager.getIntPref("export_end_date")*100000);
	endDate = jsDateToDateTime(endDate);
	var alarmOffset = item.alarmOffset;
	
	var line = [];
	//Get the details about each occurrence between startDate and endDate
	occurrences = item.getOccurrencesBetween(startDate,endDate,{});

	for each (var occurrence in occurrences) {
		line.push(serializeEvent(occurrence, eventType, alarmOffset));
	}
	details = line.join("");
	return details;
}

function serializeEvent(item, eventType, alarmOffset) {
	var line = [];
	var encodingString = "";
	if (prefManager.getBoolPref("export_as_ANSI")) {
		encodingString = ";ENCODING=QUOTED-PRINTABLE:";
	} else {
		encodingString = ";ENCODING=QUOTED-PRINTABLE;CHARSET=UTF-8:";
	}
	
	if ((item instanceof Components.interfaces.calITodo) && (!prefManager.getBoolPref("export_tasks"))) {
		return null;
	}
	
	if ((item instanceof Components.interfaces.calIEvent) && (!prefManager.getBoolPref("export_events"))) {
		return null;
	}
	
	if (item instanceof Components.interfaces.calIEvent) {       
		line.push("BEGIN:VEVENT\r\n");
		//All day events have only a date attached to them, so if there is no time attached we have to insert one
		//or the resulting file will fail to import into some programs. Hopefully, most programs  will interpret events
		//lasting from midnight to midnight as all day events. We don't use UTC time here because timezones will alter event times.
		if (item.startDate.isDate) {
			//RFC 2445 (iCalendar bible) says all day events have an inclusive start date and a non-inclusive end date.
			//Sunbird/Lightning interprets this as an event goes from day 1 to day 2 with no times attached.
			//vCalendar doesn't have anything in particular to say on the issue, but an event from day 1 at 0:00 to
			//day 1 at 24:00 seems to work for most programs..
			line.push(item.startDate ? "DTSTART:" + item.startDate.icalString + "T000000\r\n" : "");
			//We start to find problems with poorly formatted ics files which have been imported into Sunbird/Lightning
			//Some put the start date and end date as the same, which Sunbird/Lightning interpret fine, but we have to
			//allow for both situations.  So, if the end date day is the same as the start date day, don't subtract a day.
			//We're probably still screwed for a 2 day event (but that's less common, I hope)
			if (item.endDate.day == item.startDate.day) {
				endDate = item.endDate.day;
			} else {
				endDate = item.endDate.day - 1;
			}
			//Months are stored as from 0 to 11, not 1 to 12
			endMonth = item.startDate.month + 1;
			if (item.endDate) {
				line.push("DTEND:" + item.endDate.year);
				line.push(endMonth < 9 ? "0" + endMonth : endMonth);
				line.push(endDate < 9 ? "0" + endDate + "T240000\r\n" : endDate + "T240000\r\n");
			}
		} else {
			line.push(item.startDate ? "DTSTART:" + item.startDate.getInTimezone(UTC()).icalString + "\r\n" : "");
			line.push(item.endDate ? "DTEND:" + item.endDate.getInTimezone(UTC()).icalString + "\r\n" : "");
		}
		line.push(item.getProperty("STATUS") ? "STATUS:" + item.getProperty("STATUS") + "\r\n" : "");
		line.push(alarmOffset ? "DALARM:" + alarmCalc(alarmOffset, item.startDate.nativeTime) + ";;;\r\n" : "");

	} else if (item instanceof Components.interfaces.calITodo) {
		//TODOs have a different header on their entries, and also use a different property
		line.push("BEGIN:VTODO\r\n");
		//We may need to deal with all day tasks (is there such a thing?)
		line.push(item.entryDate ? "DTSTART:" + item.entryDate.getInTimezone(UTC()).icalString + "\r\n" : "");
		line.push(item.dueDate ? "DUE:" + item.dueDate.getInTimezone(UTC()).icalString + "\r\n" : "");
		line.push(item.completedDate ? "COMPLETED:" + item.completedDate.getInTimezone(UTC()).icalString + "\r\n" : "");
		//Some devices require a PRIORITY and STATUS line to be present for tasks.
		line.push(item.getProperty("PRIORITY") ? "PRIORITY:" + item.getProperty("PRIORITY") + "\r\n" : "PRIORITY:3\r\n");
		line.push(item.getProperty("STATUS") ? "STATUS:" + item.getProperty("STATUS") + "\r\n" : "STATUS:ACCEPTED\r\n");
		line.push(alarmOffset ? "DALARM:" + alarmCalc(alarmOffset, item.entryDate.nativeTime) + ";;;\r\n" : "");
	}
	
	//We use UID for individual events, and RELATED-TO for occurrences (which end up with no UID at all)
	//Should we use the same UID instead, and the start time as a recurrence Id?  Would that mess things
	//up if we don't have recurrence rules or dates defined in the parent event?
	if (eventType == "single") {
		line.push("UID:" + item.id + "\r\n");
	} else {
		if (item.recurrenceId == item.parentItem.recurrenceId) {
			line.push("UID:" + item.id + "\r\n");
		} else {
			line.push("RELATED-TO:" + item.id + "\r\n");
			//should we create a new UID in this case, or let the importing program worry about that?
		}
	}
	
	//Categories - a list separated by commas
	if (item.getCategories({})) {
		var categoriesLine = [];
		for each (category in item.getCategories({})) {
			categoriesLine.push(";" + quotedPrintableEncoder(category));
		}
		categoriesLine = categoriesLine.join("");
		//Strip out the first semi-colon
		categoriesLine = categoriesLine.substring(1);
		//Only push the line if there are actually categories
		if (categoriesLine) {
			line.push("CATEGORIES" + encodingString + categoriesLine + "\r\n");
		}
	}
	//0.9 and before used a different way to get categories, and only supported one category
	line.push(item.getProperty("CATEGORIES") ? "CATEGORIES" + encodingString + item.getProperty("CATEGORIES") + "\r\n" : "");
	
	line.push(item.title ? "SUMMARY" + encodingString + quotedPrintableEncoder(item.title) + "\r\n" : "");
	line.push(item.getProperty("DESCRIPTION") ? "DESCRIPTION" + encodingString + quotedPrintableEncoder(item.getProperty("DESCRIPTION")) + "\r\n" : "");
	line.push(item.getProperty("LOCATION") ? "LOCATION" + encodingString + quotedPrintableEncoder(item.getProperty("LOCATION")) + "\r\n" : "");
	line.push(item.getProperty("CLASS") ? "CLASS:" + item.getProperty("CLASS") + "\r\n" : "");	
	line.push("DCREATED:" + item.getProperty("CREATED").icalString + "\r\n");
	line.push("LAST-MODIFIED:" + item.getProperty("LAST-MODIFIED").icalString + "\r\n");
	line.push(item.getProperty("URL") ? "URL:" + item.getProperty("URL") + "\r\n" : "");
	
	//RELATED-TO : There are no specific relationships defined is vcs specs, so we have to lose the RELTYPE
	if (item.getProperty("RELATED-TO")) {
		for each (relation in item.getRelations({})) {
			line.push(relation.relId ? "RELATED-TO:" + relation.relId + "\r\n" : "");
		}
	}
	//Don't export recurrence rules if recurrences are set to be exported as items
	if (eventType == "single") {
		line.push(item.recurrenceInfo ? "RRULE:" + vcsRecurrenceRule(item) + "\r\n" : "");
	}

	for each (attendee in item.getAttendees({})) {
		//the properties of attendees for vcal is slightly different than for ical
		line.push("ATTENDEE");
		var role;
		var expect;
		if (attendee.role == "OPT-PARTICIPANT") {role = "ATTENDEE"; expect = "REQUEST"}
		if (attendee.role == "REQ-PARTICIPANT") {role = "ATTENDEE"; expect = "REQUIRE"}
		if (attendee.role == "CHAIR") {role = "ORGANIZER"}
		line.push(attendee.role ? ";ROLE=" + role : "");
		line.push(expect ? ";EXPECT=" + expect : "");
		line.push(attendee.participationStatus ? ";STATUS=" + attendee.participationStatus : "");
		line.push(":" + attendee.id.substring(7) + "\r\n");
	}

	if (item instanceof Components.interfaces.calIEvent) {
		line.push("END:VEVENT\r\n");
	} else if (item instanceof Components.interfaces.calITodo) {
		line.push("END:VTODO\r\n");
	} 
	line = line.join("");
	return line;
}

//Function to calculate alarm times
function alarmCalc(aAlarmOffset, aStartDate) {
	//Get the components of the alarm time as milliseconds.
	var minutes = aAlarmOffset.minutes * 60000;
	var hours = aAlarmOffset.hours * 3600000;
	var days = aAlarmOffset.days * 86400000;
	var weeks = aAlarmOffset.weeks * 604800000;

	//Add them up to get the number of milliseconds the alarm is before the start on the event
	var offset = minutes + hours + days + weeks;
	//Subtract them from the start time of the event
	var alarmTime = aStartDate/1000 - offset;
	//Convert the result into a date format and send it to a function to convert it back to an icalstring in UTC time.
	var date = new Date(alarmTime);
	var alarm = jsDateToDateTime(date);
	return alarm.getInTimezone(UTC()).icalString;
}

function vcsRecurrenceRule(aItem) {
	var recurrenceInfo = aItem.recurrenceInfo;
	var str = [];
	// Get the rules and exceptions
	var ritems = recurrenceInfo.getRecurrenceItems({});
	var rules = [];
	var exceptions = [];
	for each (var r in ritems) {
		if (r.isNegative) {
			exceptions.push(r);
		} else {
			rules.push(r);
		}
	}
	
	if (rules.length > 0) {
		var rule = rules[0];
	}

	const dayTable = {
		1 : "SU",
		2 : "MO",
		3 : "TU",
		4 : "WE",
		5 : "TH",
		6 : "FR",
		7 : "SA"
	};
	if (rule instanceof Components.interfaces.calIRecurrenceRule) {
		// There are four main types of rules, defined by their intervals.
		//Some of them have subtypes as well.
		var count;
		var endDate;

		if (rule.isByCount) {
			if (rule.count == "-1") { count = "0" } else {count = rule.count };
		} else {
			endDate = rule.endDate ? rule.endDate.getInTimezone(UTC()) : null;
			count = "0"
		}

		switch (rule.type) {
			case "DAILY":
				ruleType = "D";
				str.push(ruleType + rule.interval + " ");
				str.push(endDate ? endDate.icalString : "#" + count);
				break;
			case "WEEKLY":
				ruleType= "W";
				//Weekly recurrences can have a BYDAY component
				if (rule.getComponent("BYDAY", {})) {
					var chunk = [];
					for each (var i in rule.getComponent("BYDAY", {})) {
						//Correlate each of the days in the BYDAY part of the rule with actual days
						day = dayTable[i];
						chunk.push(day + " ");
					}
					var days = chunk.join('')+"";
				}
				str.push(ruleType + rule.interval + " ");
				str.push(days ? days : "");
				str.push(endDate ? " " + endDate.icalString : "#" + count);
				break;
			case "MONTHLY":
				ruleType = "M";
				var chunk = [];
				var chunk2 = [];
				for each (var i in rule.getComponent("BYDAY",{})) {
					//We need to find the week of the month and the day of the week here
					//First, see if the component is positive or negative (negative means "from the end of the month")
					if (i < 0) { j = i * -1; k = "-";} else {j = i; k = "+"}
					var weekDay = j % 8;
					dayName = dayTable[weekDay];
					var weekNumber = (j - weekDay ) / 8;
					chunk.push(weekNumber + k + " " + dayName +" ");
				}
				var byDay = chunk.join('')+"";
				
				//BYMONTHDAY rules need a day of the month to repeat on
				for each (var i in rule.getComponent("BYMONTHDAY",{})) {
					//We can have the last day of the month, a day number
					//or number of days from the end of the month.
					
					if (i == "-1") { 
						chunk2.push("LD"); 
					} else if (i < "-1") {
						chunk2.push(i + "- ");
					} else if (i > "0") {
						chunk2.push(i + " ");
					}
				}
				var byMonthDay = chunk2.join('') + "";

				//If the rule has neither, then it still needs a monthday to repeat on
				if (!byMonthDay && !byDay) {
					byMonthDay = item.startDate.icalString;
					byMonthDay = byMonthDay.substring(6,8) + " ";
					if (byMonthDay.charAt(0) == "0") { 
						byMonthDay = byMonthDay.substring(1) 
					}
				}
				
				if (byDay) { 
					subType = "P"; 
				} else { 
					subType = "D"; 
				}
				
				//first, the MP or MD part (the interval goes with it)
				str.push(ruleType + subType + rule.interval + " ");
				
				//Then the BYDAY or BYMONTHDAY bits (or the day of the month that the start day is on)
				str.push(byDay ? byDay : "");
				str.push(byMonthDay ? byMonthDay : "");
				
				//Then the no of repeats or end date
				//str.push(count ? "#" + count : "" );
				str.push(endDate ? " " + endDate.icalString : "#" + count);
				break;
			case "YEARLY":
				ruleType = "YM";
				//There are subrules for years too, but Sunbird and Lightning don't care about them
				var day = item.startDate.icalString;
				day = day.substring(4,6) + " ";
				if (day.charAt(0) == "0") { day = day.substring(1) }
				
				str.push(ruleType + rule.interval + " " + day);
				str.push(endDate ? " " + endDate.icalString : "#" + count);
				break;
			default:
				//We need some sort of default case
				ruleType = "None";
				break;
		}
	}

	recur = str.join('')+"";
	recur = recur.replace(/  / , " ");
	return recur;
}

//Helper functions duplicated from other files. Prefer to use them direct from other files instead if possible.
function jsDateToDateTime(aDate) {
	var newDate = Components.classes["@mozilla.org/calendar/datetime;1"].createInstance(Components.interfaces.calIDateTime);
	newDate.jsDate = aDate;
	return newDate;
}

function quotedPrintableEncoder(str) {
	if (prefManager.getBoolPref("export_as_ANSI")) {
		//ANSI encoding
		//Need to replace = as it can cause problems later
		str = str.replace(/=/g, "!3D!")
		for (i = 0; i < str.length; i++) {
			if (!(str.charCodeAt(i) >= "33" && str.charCodeAt(i) <= "60") && !(str.charCodeAt(i) >= "62" && str.charCodeAt(i) <= "126")) {
				d = str.charCodeAt(i);
				if (d < 15) { 
					d = "=0" + d.toString(16);
				} else {
					d = "=" + d.toString(16);
				}
				d = d.toUpperCase();
				str = str.replace(str.charAt(i), d);
				i = i+2;
			}
		}
		//Replace the = as =3D from earlier
		str = str.replace(/=213D=21/g, "=3D");
		while (str.match("=0A0A")) {
			str = str.replace("=0A0A", "=0A=0A");
		}
		str = str.replace(/=0A/g, "=0D=0A=\r\n  ");
		str = str.replace(/=20/g, " ");
		return str;
	} else {
		//UTF-8 encoding
		str = encodeURIComponent(str);
		while (str.match("%0A0A")) {
			str = str.replace("%0A0A", "%0A%0A");
		}
		str = str.replace(/%20/g, " ");
		str = str.replace(/%/g, "=");
		str = str.replace(/=0A/g, "=0D=0A=\r\n  ");
		return str;
	}
}

function getTimezoneService() {
	if (getTimezoneService.mObject === undefined) {
		getTimezoneService.mObject = Components.classes["@mozilla.org/calendar/timezone-service;1"].getService(Components.interfaces.calITimezoneService);
	}
	return getTimezoneService.mObject;
}

function UTC() {
	if (UTC.mObject === undefined) {
		UTC.mObject = getTimezoneService().UTC;
	}
	return UTC.mObject;
}

function stringName(aName) {
	var gBundle = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
	var mystrings = gBundle.createBundle("chrome://vcssupport/locale/vcssupport.properties");
	return mystrings.GetStringFromName(aName);
}
