var gBundle = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
var mystrings = gBundle.createBundle("chrome://vcssupport/locale/vcssupport.properties");
var vcsFileType = mystrings.GetStringFromName("vcsFileType");
var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);

function calVcsImporter() {
	this.wrappedJSObject = this;
}

calVcsImporter.prototype.QueryInterface = function QueryInterface(aIID) {
	if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.calIImporter)) {
		throw Components.results.NS_ERROR_NO_INTERFACE;
	}

	return this;
};

calVcsImporter.prototype.getFileTypes = function getVcsFileTypes(aCount) {
	aCount.value = 1;
	return([{defaultExtension:'vcs', extensionFilter:'*.vcs', description:vcsFileType}]);
};

calVcsImporter.prototype.importFromStream = function openVcs(inputStream, aCount) {
		const nsIScriptableInputStream = Components.interfaces.nsIScriptableInputStream;

		var scriptableInputStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(nsIScriptableInputStream);

		scriptableInputStream.init(inputStream);
		var stream = scriptableInputStream.read(-1);
		scriptableInputStream.close;
		
		var str = vcsImportReplace(stream);
		var parser = Components.classes["@mozilla.org/calendar/ics-parser;1"].createInstance(Components.interfaces.calIIcsParser);
		parser.parseString(str, null);
		return parser.getItems(aCount);
}

function vcsImportReplace (str) {
	
	var line = [];
	//chop str up into individual lines to help prevent accidentally replacing stuff that we don't want to
	var time = 0;
	//Replace the quoted printable code for a new line with a newline character (MS Works uses =0D=0A= rather than =0D=0A for some reason)
	str = str.replace(/DESCRIPTION;ENCODING=QUOTED-PRINTABLE:=\r\n/ , "DESCRIPTION;ENCODING=QUOTED-PRINTABLE:");
	str = str.replace(/=0D=0A=\r\n/g, "=0D=0A");
	str = str.replace(/=0D=0A=\n/g, "=0D=0A");
	str = str.replace(/=\r\n/g, "")
	while(str.match("\n")) {
		var place = str.indexOf("\n");
		var newstr = str.substring(0,place+1);
		
		if (newstr.match(";ENCODING=QUOTED-PRINTABLE")) { 
			newstr = newstr.replace(";ENCODING=QUOTED-PRINTABLE", "");
			newstr = quotedPrintableDecoder(newstr);
		}
		
		//Change the version line to the correct event
		newstr = newstr.replace("VERSION:1.0" , "VERSION:2.0")
		
		//   ***   MS Works specific stuff  ***
		//MS Works uses the wrong version numver in cal files if produces
		newstr = newstr.replace("VERSION:5.0" , "VERSION:2.0");
		//Sunbird and Lightning need all caps in these fields - MS Works uses caps and small letters
		newstr = newstr.replace("BEGIN:vCalendar" , "BEGIN:VCALENDAR");
		newstr = newstr.replace(/BEGIN:vEvent/g , "BEGIN:VEVENT");
		newstr = newstr.replace("END:vCalendar" , "END:VCALENDAR");
		newstr = newstr.replace(/END:vEvent/g , "END:VEVENT");
		
		newstr = newstr.replace("DCREATED:" , "CREATED:");
		
		//This is to give the alarm a title
		if (newstr.match("SUMMARY:")) {
			var summary = newstr.substring(8);
		}	
		
		if (newstr.match("CATEGORIES")) {
			newstr = newstr.replace(/;/g, ",");
			if (newstr.match("CHARSET=UTF-8")){
				newstr = newstr.replace(",", ";");
			}
		}
		
		newstr = newstr.replace("RRULE: ", "RRULE:");
		
		if (newstr.match("RRULE:")) {
			//First, look for ocurrences of weekdays
			if (newstr.match(/SU|MO|TU|WE|TH|FR|SA/)) {
				newstr = newstr.replace(/\s/, ";BYDAY=");
			} 
			
			//Then look at the daily, weekly, monthly, yearly components
			if (newstr.match(/D\d/)) {
				var interval = newstr.charAt(newstr.indexOf("D") + 1);
				newstr = newstr.replace(/D\d/ , "FREQ=DAILY;INTERVAL=" + interval);
			} else if (newstr.match(/W\d/)) {
				interval = newstr.charAt(newstr.indexOf("W") + 1);
				newstr = newstr.replace(/W\d/ , "FREQ=WEEKLY;INTERVAL=" + interval);
			} else if (newstr.match("MP")) {
				newstr = newstr.replace("MP", "FREQ=MONTHLY;INTERVAL=");
				newstr = newstr.replace(/\s/, ";BYMONTHDAY=");
			} else if (newstr.match("MD")) {
				newstr = newstr.replace("MD", "FREQ=MONTHLY;INTERVAL=");
				newstr = newstr.replace(/\s/, ";BYDAY=");
			} else if (newstr.match("YM")) {
				newstr = newstr.replace("YM", "FREQ=YEARLY;INTERVAL=");
				newstr = newstr.replace(/\s/, ";BYMONTH=");
				//newstr = newstr.replace(/\s/g, ",");
			} else if (newstr.match("YD")) {
				newstr = newstr.replace("YD", "FREQ=YEARLY;INTERVAL=");
				newstr = newstr.replace(/\s/, ";BYYEARDAY=");
				//newstr = newstr.replace(/\s/g, ",");
			}
			//Misc Stuff
			
			//An 8 figure number means there is a date.
			var position = newstr.indexOf(newstr.match(/\d\d\d\d\d\d\d\d/));
			var until = newstr.substring(position, position + 8);
			if (str.match(/\d\d\d\d\d\d\d\dT/)) {
				newstr = newstr.replace(/\d\d\d\d\d\d\d\d/, ";UNTIL=" + until);
			} else {
				newstr = newstr.replace(/\d\d\d\d\d\d\d\d/, ";UNTIL=" + until + "T235959Z");
			}
			newstr = newstr.replace(/#/ , ";COUNT=");
			newstr = newstr.replace(/  / , " ");
			newstr = newstr.replace(/; / , ";");
			newstr = newstr.replace(/ ;/ , ";");
			newstr = newstr.replace(/;;/ , ";");
			newstr = newstr.replace(/,,/ , ",");
			newstr = newstr.replace(/ /g , ",");
			newstr = newstr.replace(";COUNT=0", "");
		}
		
		if (newstr.match("ATTENDEE")) {
			//We need to put attendees stuff here.  Just deleting them for now
			newstr = "";
		}
		
		//RELATED-TO Property has no relatioship type in vcs.  Lets assume that they are all siblings
		newstr = newstr.replace("RELATED-TO:", "RELATED-TO;RELTYPE=SIBLING:");
		
		//We need the start time to calculate an alarm offset (since for now, Sunbird/Lightning doesn't support absolute alarm times
		if (newstr.match("DTSTART:")) {
			//Get the start time.  It messes up events early in the morning, but it's better than nothing
			var start = newstr.substring(newstr.indexOf(":") + 10, newstr.indexOf(":") + 14)
			hours = start.substring(0,2) * 1;  //Gives us hours
			mins = start.substring(3,4) * 1; //Gives us minutes
			time = (hours * 60) + mins;  //Gives us the time as minutes
		}

		if (newstr.match("DALARM:")) {
			newstr = newstr.replace("\n" , "\nDESCRIPTION:" + summary +"\nEND:VALARM\n");
			//Hack -  messes up if the alarm time is not on the same day
			//var alarmStart = newstr.substring(newstr.indexOf("T") + 1, newstr.indexOf("T") + 5);
			
			var alarmStart = newstr.substring(newstr.indexOf("T") - 8, newstr.indexOf("T") + 5);

			alarmYears = alarmStart.substring(0,3);
			alarmMonths = alarmYears.substring(4,5);
			alarmDays = alarmStart.substring(6,7);
			alarmHours = alarmStart.substring(9,10) * 1;
			alarmMins = alarmStart.substring(11,12) * 1;
			alarm = (alarmHours * 60) + alarmMins;
			offset = time - alarm;
			newstr = newstr.replace(/DALARM:\d\d\d\d\d\d\d\dT\d\d\d\d\d\dZ/,"BEGIN:VALARM\nTRIGGER;VALUE=DURATION:-PT" + offset + "M");
			//Preferred method
			//newstr = newstr.replace("DALARM:" , "BEGIN:VALARM\nTRIGGER;VALUE=DATE-TIME:");
			newstr = newstr.replace(";;;" , "")
		}
		
		//Remove any spare spaces from date lines (see bug 421875)
		if (newstr.match("DTSTART" || "DTEND" || "DALARM" || "DTSTAMP" || "LAST-MODIFIED" || "DCREATED")) {
			newstr = newstr.replace(/ /g , "");	
		}

		newstr = newstr.replace(/,;/ , ";");
		line.push(newstr);
		str=str.substring(place+1);
	}
	//Add the last line in
	line.push(str);
	//Put the lines back together again
	line = line.join("");
	//Replace any extra newlines added with a single newline
	line = line.replace(/\n\n/ , "\n");
	//promptService.alert(null, "Title", line);
	return line;
}

function quotedPrintableDecoder(str) {
	
	//=3D maps to = in Quoted printable, which will give us loop issues.  Change it for now
	str = str.replace(/=3D/g, "!3D!");
	str = str.replace(/=0D/g , "");
	while (str.match("=0A=0A")) {
		str = str.replace(/=0A=0A/g, "\\n\\n");
	}
	str = str.replace(/=0A=/g, "\\n");
	str = str.replace(/=0A/g , "\\n");
	str = str.replace(/\\n  /g, "\\n");
	str = str.replace(";ENCODING=QUOTED-PRINTABLE:", ":");
	var start = str.indexOf(":");
	while (str.match("=")){
		var match = str.indexOf("=");
		if (match < start){
			str = str.replace("=","!3D!");
		} else {
			match = str.indexOf("=");
			if (str.match(";CHARSET!3D!UTF-8")) {
				str1 = str.substring(0, str.indexOf(":")+1);
				str2 = str.substring(str.indexOf(":")+1, str.length);
				str2 = str2.replace(/=/g, "%");
				str2 = unescape(decodeURIComponent(str2));
				str2 = str2.replace(/%/g, "=");	
				str = str1 + str2;
			} else {
				var test = str.substring(match+1, match+3);
				test = parseInt(test, 16);
				test = test.toString(10);
				str = str.replace(str.charAt(match), String.fromCharCode(test));
				//We need to replace the next 2 letters after the =
				str1 = str.substring(0,match+1);
				str2 = str.substring(match+3, str.length);
				str = str1 + str2;
			}
		}
	}
	//Put back in the =
	str = str.replace(/!3D!/g, "=");
	return str;
}
